from __future__ import annotations

from os.path import dirname, join
from typing import Optional

import pandas as pd


def _load_dataset(
    file_name: str,
    colnames: Optional[list],
    name: str,
    description: str,
    design: dict,
    source: str,
) -> None:

    module_path = dirname(__file__)
    file_path = join(module_path, "data", file_name)
    df = pd.read_csv(file_path)
    if colnames is not None:
        df = df[colnames]
    nrows, ncols = df.shape

    return {
        "name": name,
        "description": description,
        "nrows": nrows,
        "ncols": ncols,
        "data": df,
        "design": design,
        "source": source,
    }


def load_psu_frame():
    name = "PSU Frame"
    description = "A simulated census data."
    design = {}
    source = ""

    return _load_dataset(
        "psu_frame.csv",
        colnames=None,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_psu_sample():
    colnames = ["cluster", "region", "psu_prob"]
    name = "PSU Sample"
    description = "The PSU sample obtained from the simulated PSU frame."
    design = {}
    source = ""

    return _load_dataset(
        "psu_sample.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_ssu_sample():
    colnames = ["cluster", "household", "ssu_prob"]
    name = "SSU Sample"
    description = "The SSU sample obtained from the simulated SSU frame."
    design = {}
    source = ""

    return _load_dataset(
        "ssu_sample.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_nhanes2():
    colnames = [
        "stratid",
        "psuid",
        "race",
        "highbp",
        "highlead",
        "zinc",
        "diabetes",
        "finalwgt",
    ]
    name = "NHANES II Subsample"
    description = "A subset of NHANES II data. This file is not meant to be representative of NHANES II. It is just an subset to illustrate the syntax in this tutorial."
    design = {}
    source = ""

    return _load_dataset(
        "nhanes2.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_nhanes2brr():
    colnames = None
    name = "NHANES II Subsample with bootstrap weights"
    description = "A subset of NHANES II data with bootstrap weights. This file is not meant to be representative of NHANES II. It is just an subset to illustrate the syntax in this tutorial."
    design = {}
    source = ""

    return _load_dataset(
        "nhanes2brr_subset.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_nhanes2jk():
    colnames = None
    name = "NHANES II Subsample with jackknife weights"
    description = "A subset of NHANES II data with jackknife weights. This file is not meant to be representative of NHANES II. It is just an subset to illustrate the syntax in this tutorial."
    design = {}
    source = ""

    return _load_dataset(
        "nhanes2jk_subset.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_nmihs():
    colnames = None
    name = "NMIHS Subsample"
    description = "A subset of nmihs data. This file is not meant to be representative of nmihs. It is just an subset to illustrate the syntax in this tutorial."
    design = {}
    source = ""

    return _load_dataset(
        "nmihs_subset.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_auto():
    colnames = None
    name = "Auto Sample"
    description = "The Auto sample data."
    design = {}
    source = ""

    return _load_dataset(
        "auto.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_birth():
    colnames = None
    name = "Birth Sample"
    description = "The Birth sample data."
    design = {}
    source = ""

    return _load_dataset(
        "birth.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_county_crop():
    colnames = [
        "county_id",
        "corn_area",
        "soybeans_area",
        "corn_pixel",
        "soybeans_pixel",
    ]
    name = "County Crop Sample"
    description = "The County Crop Areas sample data."
    design = {}
    source = ""

    return _load_dataset(
        "countycrop.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_county_crop_means():
    colnames = [
        "county_id",
        "samp_segments",
        "pop_segments",
        "ave_corn_pixel",
        "ave_soybeans_pixel",
    ]
    name = "County Crop Area Means"
    description = "The County Crop Area Means data."
    design = {}
    source = ""

    return _load_dataset(
        "countycrop_means.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )


def load_expenditure_milk():
    colnames = [
        "major_area",
        "small_area",
        "samp_size",
        "direct_est",
        "std_error",
        "coef_var",
    ]
    name = "Expenditure on Milk"
    description = "The expenditure on milk data."
    design = {}
    source = ""

    return _load_dataset(
        "expenditure_on_milk.csv",
        colnames=colnames,
        name=name,
        description=description,
        design=design,
        source=source,
    )



from enum import Enum, unique




@unique
class FitMethod:
    ols= "OLS"
    wls= "WLS"
    fh = "FH"
    ml= "ML"
    reml= "REML"

@unique
class Mse(Enum):
    taylor = "Taylor"
    boot = "Boostrap"
    jkn = "Jackknife"


# Population parameters
@unique
class PopParam(Enum):
    count = "Count"
    mean = "Mean"
    total = "Total"
    prop = "Proportion"
    ratio = "Ratio"


@unique
class RepMethod(Enum):
    jackknife  = "Jackknife"
    bootstrap = "Bootstrap"
    brr = "BRR"
    
    
# Methods for sample size
@unique
class SizeMethod(Enum):
    wald = "Wald"
    fleiss = "Fleiss"


@unique
class SelectMethod(Enum):
    srs_wr = "SRS with replacement"
    srs_wor = "SRS without replacement"
    sys = "Systematic"
    pps_brewer = "PPS Brewer"
    pps_hv = "PPS Hanurav-Vijayan"
    pps_murphy = "PPS Murphy"
    pps_rs = "PPS Rao-Sampford"
    pps_sys = "PPS Systematic"
    pps_wr = "PPS with replacement"
    grs = "General"


@unique
class SinglePSUEst(Enum):
    """Estimation options for strata with singleton PSU"""

    error = "Raise Error when one PSU in a stratum"
    skip = "Set variance to zero and skip stratum with one PSU"
    certainty = "Use SSUs or lowest units to estimate the variance"
    combine = "Combine the strata with the singleton psu to another stratum"



import pandas as pd
import polars as pl

from samplics.types.basic import DataFrame
from samplics.types.protocols import SamplePrcl


class Sample(SamplePrcl):
    data: dict
    schema: dict

    __slots__ = ["data", "schema"]

    def __init__(self, data: dict | DataFrame, schema: dict):
        match data:
            case dict():
                self.data = data
            case pd.DataFrame:
                self.data = data.to_dict()
            case pl.DataFrame:
                self.data = data.to_dict()
            case _:
                raise TypeError("data must be a dictionary or a DataFrame!")

        self.schema = schema

    def estimate(self):
        pass

    def adjust(self):
        pass



import math

from typing import Any, Optional, Tuple, TypeVar, Union

import numpy as np
import pandas as pd

from scipy.stats import t as student

from samplics.estimation.expansion import _SurveyEstimator
from samplics.utils.formats import numpy_array, remove_nans
from samplics.utils.types import Array, Number, PopParam, RepMethod


TypeRepEst = TypeVar("TypeRepEst", bound="ReplicateEstimator")


class ReplicateEstimator(_SurveyEstimator):
    """*ReplicateEstimator* implements replicate-based variance approximations.


    Attributes
        | point_est (dict): point estimate of the parameter of interest.
        | variance (dict): variance estimate of the parameter of interest.
        | stderror (dict): standard error of the parameter of interest.
        | coef_var (dict): estimate of the coefficient of variation.
        | deff (dict): estimate of the design effect due to weighting.
        | lower_ci (dict): estimate of the lower bound of the confidence interval.
        | upper_ci (dict): estimate of the upper bound of the confidence interval.
        | degree_of_freedom (int): degree of freedom for the confidence interval.
        | alpha (float): significant level for the confidence interval
        | strata (list): list of the strata in the sample.
        | domains (list): list of the domains in the sample.
        | method (str): variance estimation method.
        | param (str): the parameter of the population to estimate e.g. total.
        | nb_strata (int): number of strata.
        | nb_psus (int): number of primary sampling units (psus).
        | conservative (bool): indicate whether to produce conservative variance estimates.
        | nb_reps (int): number of replicate weights.
        | rep_coefs (array): coefficients associated to the replicate weights.
        | fay_coef (float): Fay coefficient for the the BRR-Fay algorithm.
        | rand_seed (int): random seed for reproducibility.

    Methods
        | estimate(): produces the point estimate of the parameter of interest with the associated
        |   measures of precision.
    """

    def __init__(
        self,
        method: str,
        param: str,
        rep_weight_cls: Optional[Any] = None,
        fay_coef: Optional[Number] = None,
        alpha: float = 0.05,
        rand_seed: Optional[int] = None,
    ) -> None:

        if method not in (RepMethod.bootstrap, RepMethod.brr, RepMethod.jackknife):
            raise ValueError("Method must be 'bootstrap', 'brr', or 'jackknife'!")

        super().__init__(param, alpha, rand_seed)
        self.method = method
        self.conservative = False
        self.degree_of_freedom: Optional[int] = None
        self.nb_reps: Optional[int] = None
        self.rep_coefs: Optional[Union[np.ndarray, Number]] = None
        if method == RepMethod.brr and fay_coef is not None:
            self.fay_coef = fay_coef
        elif method == RepMethod.brr:
            self.fay_coef = 0
        if rep_weight_cls is not None:
            self.nb_reps = rep_weight_cls.nb_reps
            self.rep_coefs = rep_weight_cls.rep_coefs
            self.degree_of_freedom = rep_weight_cls.degree_of_freedom
            self.fay_coef = rep_weight_cls.fay_coef if self.method == RepMethod.brr else None

    def _rep_point(
        self, y: np.ndarray, rep_weights: np.ndarray, x: Optional[np.ndarray]
    ) -> np.ndarray:
        if self.param in (PopParam.prop, PopParam.mean):
            return np.asarray(
                np.sum(rep_weights * y[:, None], axis=0) / np.sum(rep_weights, axis=0)
            )
        elif self.param == PopParam.total:
            return np.asarray(np.sum(rep_weights * y[:, None], axis=0))
        elif self.param == PopParam.ratio and x is not None:
            return np.asarray(
                np.sum(rep_weights * y[:, None], axis=0)
                / np.sum(rep_weights * x[:, None], axis=0)
            )
        else:
            raise AssertionError("Parameter not valid!")

    def _bias(
        self,
        y: np.ndarray,
        samp_weight: np.ndarray,
        rep_weights: np.ndarray,
        x: Optional[np.ndarray],
    ) -> Number:

        estimate = self._get_point(y, samp_weight, x)
        if isinstance(estimate, (int, float)):
            rep_estimates = self._rep_point(y, rep_weights, x)
            return float(np.sum(np.mean(rep_estimates) - estimate))
        else:
            raise AssertionError

    def _rep_coefs(self, rep_coefs: Optional[Union[np.ndarray, Number]] = None) -> None:

        if rep_coefs is not None:
            if isinstance(rep_coefs, np.ndarray):
                self.rep_coefs = rep_coefs
            elif isinstance(rep_coefs, (int, float)):
                self.rep_coefs = rep_coefs
        elif self.nb_reps is not None and self.method == RepMethod.bootstrap:
            self.rep_coefs = (1 / self.nb_reps) * np.ones(self.nb_reps)
        elif self.nb_reps is not None and self.method == RepMethod.brr:
            self.rep_coefs = (1 / (self.nb_reps * pow(1 - self.fay_coef, 2))) * np.ones(
                self.nb_reps
            )
        elif self.nb_reps is not None and self.method == RepMethod.jackknife:
            self.rep_coefs = ((self.nb_reps - 1) / self.nb_reps) * np.ones(self.nb_reps)

    def _variance(
        self,
        y: np.ndarray,
        rep_weights: np.ndarray,
        rep_coefs: np.ndarray,
        x: np.ndarray,
        estimate: float,
        conservative: bool = False,
    ) -> Number:

        variance = 0.0
        rep_estimates = self._rep_point(y, rep_weights, x)
        if self.method == RepMethod.jackknife:  # page 155 (4.2.3 and 4.2.5) - Wolter(2003)
            jk_factor = np.array(1 / (1 - rep_coefs))
            pseudo_estimates = jk_factor * estimate - (jk_factor - 1) * rep_estimates
            if conservative:
                variance = float(
                    np.sum(
                        rep_coefs
                        * pow((pseudo_estimates - estimate) / (jk_factor - 1), 2)
                    )
                )
            elif not conservative:
                variance = float(
                    np.sum(
                        rep_coefs
                        * pow(
                            (pseudo_estimates - np.mean(pseudo_estimates))
                            / (jk_factor - 1),
                            2,
                        )
                    )
                )
        else:
            if conservative:
                variance = float(np.sum(rep_coefs * pow(rep_estimates - estimate, 2)))
            elif not conservative:
                variance = float(
                    np.sum(rep_coefs * pow(rep_estimates - np.mean(rep_estimates), 2))
                )
        return variance

    def _get_bias(
        self,
        y: np.ndarray,
        samp_weight: np.ndarray,
        rep_weights: np.ndarray,
        x: Optional[np.ndarray] = None,
        domain: Optional[np.ndarray] = None,
        remove_nan: bool = False,
    ) -> Any:

        if remove_nan:
            if self.param == PopParam.ratio and x is not None:
                excluded_units = np.isnan(y) | np.isnan(x)
            else:
                excluded_units = np.isnan(y)
            y, samp_weight, x, stratum, domain, _, _ = remove_nans(
                excluded_units, y, samp_weight, x, None, domain, None, None
            )
            rep_weights = rep_weights[~excluded_units, :]

        if self.param == PopParam.prop:
            y_dummies = pd.get_dummies(y)
            categories = y_dummies.columns
            y_dummies = y_dummies.values

        bias: Any
        if domain is None:
            if self.param == PopParam.prop and categories is not None:
                cat_dict = dict()
                for k in range(categories.size):
                    cat_dict_k = dict(
                        {
                            categories[k]: self._bias(
                                y_dummies[:, k], samp_weight, rep_weights, None
                            )
                        }
                    )
                    cat_dict.update(cat_dict_k)
                bias = cat_dict
            else:
                bias = self._bias(y, samp_weight, rep_weights, x)
        else:
            for d in np.unique(domain):
                samp_weight_d = samp_weight * (domain == d)
                rep_weights_d = rep_weights * (domain == d)[:, None]
                if self.param == PopParam.ratio:
                    x_d = x * (domain == d)
                else:
                    x_d = x
                if self.param == PopParam.prop:
                    y_dummies_d = y_dummies * (domain == d)[:, None]
                    cat_dict = dict()
                    for k in range(categories.size):
                        cat_dict_d_k = dict(
                            {
                                categories[k]: self._bias(
                                    y_dummies_d[:, k],
                                    samp_weight_d,
                                    rep_weights_d,
                                    None,
                                )
                            }
                        )
                        cat_dict.update(cat_dict_d_k)
                    bias[d] = cat_dict
                else:
                    y_d = y * (domain == d)
                    bias[d] = self._bias(y_d, samp_weight_d, rep_weights, x_d)

        return bias

    def _get_variance(
        self,
        y: np.ndarray,
        samp_weight: np.ndarray,
        rep_weights: np.ndarray,
        rep_coefs: np.ndarray,
        x: Optional[np.ndarray] = None,
        domain: Optional[np.ndarray] = None,
        conservative: bool = False,
        remove_nan: bool = False,
    ) -> Any:

        if self.param == PopParam.prop:
            y_dummies = pd.get_dummies(y)
            categories = y_dummies.columns
            y_dummies = y_dummies.values

        if domain.shape in ((), (0,)):
            if self.param == PopParam.prop:
                estimate_k: dict
                cat_dict = dict()
                for k in range(categories.size):
                    estimate_k = self._get_point(
                        y=y_dummies[:, k],
                        samp_weight=samp_weight,
                        x=x,
                        domain=np.array(None),
                    )
                    cat_dict_k = dict(
                        {
                            categories[k]: self._variance(
                                y_dummies[:, k],
                                rep_weights,
                                rep_coefs,
                                x,
                                estimate_k[1],
                                conservative,
                            )
                        }
                    )
                    cat_dict.update(cat_dict_k)
                return cat_dict
            else:
                estimate = self._get_point(
                    y=y, samp_weight=samp_weight, x=x, domain=domain
                )
                return self._variance(
                    y, rep_weights, rep_coefs, x, estimate, conservative
                )
        else:
            variance_else1 = {}
            variance_else2 = {}
            for d in np.unique(domain):
                samp_weight_d = samp_weight * (domain == d)
                rep_weights_d = rep_weights * (domain == d)[:, None]
                if self.param == PopParam.ratio:
                    x_d = x * (domain == d)
                else:
                    x_d = x
                if self.param == PopParam.prop and categories is not None:
                    y_dummies_d = np.asarray(y_dummies * (domain == d)[:, None])
                    cat_dict = dict()
                    for k in range(categories.size):
                        estimate_d_k = self._get_point(
                            y=y_dummies_d[:, k],
                            samp_weight=samp_weight_d,
                            x=x_d,
                            domain=np.array(None),
                        )[1]
                        cat_dict_d_k = dict(
                            {
                                categories[k]: self._variance(
                                    y=y_dummies_d[:, k],
                                    rep_weights=rep_weights_d,
                                    rep_coefs=rep_coefs,
                                    x=x_d,
                                    estimate=estimate_d_k,
                                    conservative=conservative,
                                )
                            }
                        )
                        cat_dict.update(cat_dict_d_k)
                    variance_else1[d] = cat_dict
                else:
                    y_d = y * (domain == d)
                    estimate_d = self._get_point(
                        y=y_d, samp_weight=samp_weight_d, x=x_d, domain=np.array(None)
                    )
                    variance_else2[d] = self._variance(
                        y=y_d,
                        rep_weights=rep_weights_d,
                        rep_coefs=rep_coefs,
                        x=x_d,
                        estimate=estimate_d,
                        conservative=conservative,
                    )
            if self.param == PopParam.prop:
                return variance_else1
            else:
                return variance_else2

    def _get_confint(
        self,
        param: str,
        estimate: Any,
        variance: Any,
        quantile: float,
    ) -> Tuple[Any, Any]:

        lower_ci = {}
        upper_ci = {}
        if self.domains.shape in ((), (0,)):
            if param == PopParam.prop:
                for level in variance:
                    point_est = estimate[level]
                    std_est = pow(variance[level], 0.5)
                    location_ci = math.log(point_est / (1 - point_est))
                    scale_ci = std_est / (point_est * (1 - point_est))
                    ll = location_ci - quantile * scale_ci
                    lower_ci[level] = math.exp(ll) / (1 + math.exp(ll))
                    uu = location_ci + quantile * scale_ci
                    upper_ci[level] = math.exp(uu) / (1 + math.exp(uu))
                return lower_ci, upper_ci
            else:
                return estimate - quantile * pow(
                    variance, 0.5
                ), estimate + quantile * pow(variance, 0.5)

        else:
            lower_ci_else1 = {}
            upper_ci_else1 = {}
            lower_ci_else2 = {}
            upper_ci_else2 = {}
            for key in variance:
                if param == PopParam.prop:
                    lower_ci_k = {}
                    upper_ci_k = {}
                    for level in variance[key]:
                        point_est = estimate[key][level]
                        std_est = pow(variance[key][level], 0.5)
                        location_ci = math.log(point_est / (1 - point_est))
                        scale_ci = std_est / (point_est * (1 - point_est))
                        ll = location_ci - quantile * scale_ci
                        lower_ci_k[level] = math.exp(ll) / (1 + math.exp(ll))
                        uu = location_ci + quantile * scale_ci
                        upper_ci_k[level] = math.exp(uu) / (1 + math.exp(uu))
                    lower_ci_else1[key] = lower_ci_k
                    upper_ci_else1[key] = upper_ci_k
                else:
                    lower_ci_else2[key] = estimate[key] - quantile * pow(
                        variance[key], 0.5
                    )
                    upper_ci_else2[key] = estimate[key] + quantile * pow(
                        variance[key], 0.5
                    )

            if self.param == PopParam.prop:
                return lower_ci_else1, upper_ci_else1
            else:
                return lower_ci_else2, upper_ci_else2

    def _get_coefvar(
        self,
        param: str,
        estimate: Any,  # Any := Union[dict[StringNumber, DictStrNum], DictStrNum, Number]
        variance: Any,
    ) -> Any:  # Any := Union[dict[StringNumber, DictStrNum], DictStrNum, Number]

        if self.domains.shape in ((), (0,)):
            if param == PopParam.prop:
                coef_var = {}
                for level in variance:
                    coef_var[level] = pow(variance[level], 0.5) / estimate[level]
            else:
                coef_var = pow(variance, 0.5) / estimate
        else:
            coef_var = {}
            for key in variance:
                if param == PopParam.prop:
                    coef_var_k = {}
                    for level in variance[key]:
                        coef_var_k[level] = (
                            pow(variance[key][level], 0.5) / estimate[key][level]
                        )
                    coef_var[key] = coef_var_k
                else:
                    coef_var[key] = pow(variance[key], 0.5) / estimate[key]
        return coef_var

    def estimate(
        self: TypeRepEst,
        y: Array,
        samp_weight: Array,
        rep_weights: Union[np.ndarray, pd.DataFrame],
        x: Optional[Union[np.ndarray, pd.DataFrame]] = None,
        rep_coefs: Optional[Union[float, np.ndarray]] = None,
        domain: Optional[np.ndarray] = None,
        conservative: bool = False,
        deff: bool = False,  # Todo
        remove_nan: bool = False,
    ) -> TypeRepEst:
        """[summary]

        Args:
            self (TypeRepEst): [description]
            y (Array): [description]
            samp_weight (Array): [description]
            rep_weights (Union[np.ndarray, pd.DataFrame]): [description]
            x (Union[np.ndarray, pd.DataFrame, None], optional): [description]. Defaults to None.
            rep_coefs (Union[float, np.ndarray, None], optional): [description]. Defaults to None.
            domain (Optional[np.ndarray], optional): [description]. Defaults to None.
            conservative (bool, optional): [description]. Defaults to False.
            deff (bool, optional): [description]. Defaults to False.

        Raises:
            AssertionError: [description]

        Returns:
            TypeRepEst: [description]
        """

        if self.param == PopParam.ratio and x is None:
            raise AssertionError("x must be provided for ratio estimation.")

        _y = numpy_array(y)
        _x = numpy_array(x)
        _samp_weight = numpy_array(samp_weight)
        _rep_weights = numpy_array(rep_weights)
        _rep_coefs = numpy_array(rep_coefs)
        _domain = numpy_array(domain)

        if remove_nan:
            to_keep = remove_nans(_y.shape[0], _y)

            _y = _y[to_keep]
            _x = _x[to_keep] if _x.shape not in ((), (0,)) else _x
            _samp_weight = _samp_weight[to_keep]
            _rep_coefs = (
                _rep_coefs[to_keep]
                if _rep_coefs.shape not in ((), (0,))
                else _rep_coefs
            )
            _rep_weights = (
                _rep_weights[to_keep]
                if _rep_weights.shape not in ((), (0,))
                else _rep_weights
            )
            _domain = _domain[to_keep] if _domain.shape not in ((), (0,)) else _domain

        self.conservative = conservative

        if self.nb_reps is None:
            self.nb_reps = rep_weights.shape[1]

        self._rep_coefs(rep_coefs)

        self.domains = (
            np.unique(_domain) if _domain.shape not in ((), (0,)) else _domain
        )

        self.point_est = self._get_point(
            y=_y, samp_weight=_samp_weight, x=_x, domain=_domain
        )
        self.variance = self._get_variance(
            y=_y,
            samp_weight=_samp_weight,
            rep_weights=_rep_weights,
            rep_coefs=np.array(self.rep_coefs),
            x=_x,
            domain=_domain,
            conservative=conservative,
            remove_nan=remove_nan,
        )

        if self.method == RepMethod.brr and self.degree_of_freedom is None:
            self.degree_of_freedom = int(self.nb_reps / 2)
        elif self.degree_of_freedom is None:
            self.degree_of_freedom = int(self.nb_reps) - 1

        t_quantile = student.ppf(1 - self.alpha / 2, df=self.degree_of_freedom)

        self.lower_ci, self.upper_ci = self._get_confint(
            self.param, self.point_est, self.variance, t_quantile
        )
        self.coef_var = self._get_coefvar(self.param, self.point_est, self.variance)

        if self.domains.shape in ((), (0,)):
            if self.param == PopParam.prop:
                for level in self.variance:
                    self.stderror[level] = pow(self.variance[level], 0.5)
            elif isinstance(self.variance, (int, float)):
                self.stderror = pow(self.variance, 0.5)
        else:
            for key in self.variance:
                if self.param == PopParam.prop:
                    stderror = {}
                    for level in self.variance[key]:
                        stderror[level] = pow(self.variance[key][level], 0.5)
                    self.stderror[key] = stderror
                else:
                    self.stderror[key] = pow(self.variance[key], 0.5)

        return self



"""Formats to tranform the data into another format.

Functions:
    | *numpy_array()* converts an array-like data to np.ndarray
    | *array_to_dict()* converts an array to a dictionary where the keys are the unique values of
    |   the array and the values of the dictionary are the counts of the array values.
    | *dataframe_to_array()* returns a pandas dataframe from an np.ndarray.
"""

from __future__ import annotations

from typing import Any, Optional, Union

import numpy as np
import pandas as pd
import polars as pl

from samplics.utils.checks import assert_not_unique
from samplics.utils.types import (
    Array,
    DictStrInt,
    DictStrNum,
    Number,
    Series,
    StringNumber,
)


def numpy_array(arr: Array) -> np.ndarray:
    if isinstance(arr, (pd.DataFrame, pl.DataFrame)):
        return arr.to_numpy()
    elif isinstance(arr, (pd.Series, pl.Series)):
        return arr.to_numpy()
    elif not isinstance(arr, np.ndarray):
        arr_np = np.asarray(arr)
        if isinstance(arr, (list, tuple)) and len(arr_np.shape) == 2:
            arr_np = np.transpose(arr_np)
        return arr_np
    else:
        return arr


def array_to_dict(arr: np.ndarray, domain: Optional[np.ndarray] = None) -> DictStrNum:
    if domain is None:
        keys, counts = np.unique(numpy_array(arr), return_counts=True)
        out_dict = dict(zip(keys, counts))
    else:
        out_dict = {}
        for d in np.unique(domain):
            arr_d = arr[domain == d]
            keys_d, counts_d = np.unique(numpy_array(arr_d), return_counts=True)
            out_dict[d] = dict(zip(keys_d, counts_d))

    return out_dict


def dataframe_to_array(df: pd.DataFrame) -> np.ndarray:
    if isinstance(df, (pd.Series, pl.Series)):
        return df.to_numpy()
    elif isinstance(df, (pl.DataFrame, pd.DataFrame)):
        if isinstance(df, pd.DataFrame):
            df = pl.from_pandas(df)
        nb_vars = df.shape[1]
        col_names = df.columns

        x_array = df.select(col_names[0]).to_series().cast(pl.Utf8).to_numpy()
        if nb_vars > 1:
            for k in range(1, nb_vars):
                x_array = (
                    x_array + "_&_" + df.select(col_names[k]).to_series().cast(pl.Utf8).to_numpy()
                )
        return x_array
    else:
        raise AssertionError("The input data is not a pandas dataframe")


def data_to_dict(
    data: Union[DictStrInt, int],
    strat: bool,
    stratum: Array,
) -> Union[DictStrInt, int]:
    if not isinstance(data, dict) and strat:
        return dict(zip(stratum, np.repeat(data, len(stratum))))
    if isinstance(data, (int, float)) and not strat:
        return data
    elif isinstance(data, dict):
        return data
    else:
        raise AssertionError


def sample_units(all_units: Array, unique: bool = True) -> np.ndarray:
    all_units = numpy_array(all_units)
    if unique:
        assert_not_unique(all_units)

    return all_units


def dict_to_dataframe(col_names: list[str], *args: Any) -> pd.DataFrame:
    if isinstance(args[0], dict):
        values_df = pd.DataFrame(columns=col_names)
        keys = list(args[0].keys())
        number_keys = len(keys)
        for k, arg in enumerate(args):
            args_keys = list(args[k].keys())
            if not isinstance(arg, dict) or (keys != args_keys) or number_keys != len(args_keys):
                raise AssertionError(
                    "All input parameters must be dictionaries with the same keys."
                )

            if isinstance(
                args[0].get(keys[0]), dict
            ):  # For the case of nested dictionaries e.g. proportion or as_factor=True
                keys_list = list()
                levels = list()
                values = list()
                for key in keys:
                    keys_list += np.repeat(key, len(list(arg[key].keys()))).tolist()
                    levels += list(arg[key].keys())
                    values += list(arg[key].values())
                if k == 0:
                    values_df.iloc[:, 1] = keys_list
                    values_df.iloc[:, 2] = levels
                values_df[col_names[k + 3]] = values
            else:
                if k == 0:
                    values_df.iloc[:, 1] = keys
                values_df[col_names[k + 2]] = arg.values()
    else:
        values_df = pd.DataFrame({args})
        values_df.insert(0, "_parameter", " ")
        values_df.columns = col_names

    return values_df


# def remove_nans(excluded_units: Array, *args: Any) -> list:

#     excluded_units = numpy_array(excluded_units)
#     vars_list = list()
#     for var in args:
#         if var is not None and len(var.shape) != 0:
#             vars_list.append(var[~excluded_units])
#         else:
#             vars_list.append(None)

#     return vars_list


def remove_nans(n: Number, *args: np.ndarray) -> list:
    excluded_units = np.zeros(n).astype(bool)

    for var in args:
        if var.shape not in ((), (0,)):
            assert n == var.shape[0]
            if var.dtype.kind in ("i", "u", "f", "c", "b", "m", "M"):
                excluded_units = excluded_units | np.isnan(var)
            else:
                var = var.astype(str)
                excluded_units = excluded_units | np.isin(var, ("nan", ""))

    return ~excluded_units


def fpc_as_dict(stratum: np.ndarray, fpc: Union[Array, Number]) -> Union[DictStrNum, Number]:
    if stratum.shape in ((), (0,)) and isinstance(fpc, (int, float)):
        return fpc
    elif stratum.shape not in ((), (0,)) and isinstance(fpc, (int, float)):
        return dict(zip(stratum, np.repeat(fpc, stratum.shape[0])))
    elif stratum.shape not in ((), (0,)) and isinstance(fpc, np.ndarray):
        return dict(zip(stratum, fpc))
    else:
        raise TypeError("stratum and fpc are not compatible!")


def convert_numbers_to_dicts(
    number_strata: Optional[int], *args: Union[DictStrNum, Number]
) -> list[DictStrNum]:
    dict_number = 0
    stratum: Optional[list[StringNumber]] = None
    for arg in args:
        if arg is not None and not isinstance(arg, (int, float, dict)):
            raise TypeError("Arguments must be of type int, float or dict!")

        if isinstance(arg, dict):
            dict_number += 1
            if dict_number == 1:
                stratum = list(arg.keys())
            elif dict_number >= 1:
                if stratum != list(arg.keys()):
                    raise AssertionError("Python dictionaries have different keys")

    if stratum is None:
        if isinstance(number_strata, (int, float)) and number_strata >= 1:
            stratum = ["_stratum_" + str(i) for i in range(1, number_strata + 1)]
        else:
            raise ValueError("Number of strata must be superior or equal to 1!")
    else:
        number_strata = len(stratum)

    list_of_dicts = list()
    for arg in args:
        if isinstance(arg, (int, float)):
            list_of_dicts.append(dict(zip(stratum, np.repeat(arg, number_strata))))
        else:
            list_of_dicts.append(arg)

    return list_of_dicts


def concatenate_series_to_str(row: Series) -> str:
    return "__by__".join([str(c) for c in row])


def numpy_to_dummies(arr: np.ndarray, vars_names: list[str]) -> np.ndarray:
    df = pd.DataFrame(arr.astype(str))
    df.columns = vars_names

    return np.asarray(pd.get_dummies(df, drop_first=True).to_numpy())



"""Sample weighting module

*SampleWeight* is the main class in this module which implements weight adjustments to account for
nonresponse, calibrate to auxiliary information, normalize weights, and trim extreme weights. Valliant, R. and Dever, J. A. (2018) [#vd2018]_ provides a step-by-step guide on calculating
sample weights.

.. [#vd2018] Valliant, R. and Dever, J. A. (2018), *Survey Weights: A Step-by-Step Guide to
   Calculation*, Stata Press.
"""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pandas as pd
import polars as pl

from samplics.utils import checks, formats
from samplics.utils.types import (
    Array,
    DictStrInt,
    DictStrNum,
    Number,
    StringNumber,
)


class SampleWeight:
    """*SampleWeight* implements several adjustments to sample weights. The class does not compute design sample weights. It is expected at this point some initial weights are
    available e.g. design sample weights or some other sample weights. Using this module,
    the user will be able to adjust sample weights to account for nonresponse, normalize
    sample weights so that they sum to some control value(s), poststratify, and calibrate
    based on auxiliary information.

    Attributes
        | adj_method (str): adjustment method. Possible values are nonresponse,
        |   normalization, poststratification, calibration.
        | nb_units (dict): number of units per domain.
        | deff_weight (dict): design effect due to unequal weights per domain.
        | adj_factor (dict): normalizing adjustment factor per domain.
        | control (dict): control values per domain.

    Methods
        | calculate_deff_weight(): computes the design effect due to weighting.
        | adjust(): adjust the sample weights to account for nonresponse.
        | normalize(): normalize the sample weights to ensure they sum to a control value.
        | poststratify(): poststratify the sample weights.
        | calib_covariables(): covert a dataframe to a tuple of an array and a dictionary.
        | The array corresponds to the calibration domains. The dictionary maps the array
        | elements with their corresponding control values.
        | calibrate(): calibrate the sample weights.

    TODO: trim(), rake()
    """

    def __init__(self) -> None:

        self.adj_method: str = ""
        self.nb_units: Union[DictStrInt, int] = {}
        self.deff_weight: Union[DictStrNum, Number] = {}
        self.adj_factor: Union[DictStrNum, Number] = {}
        self.control: Union[DictStrNum, Number] = {}

    def __repr__(self) -> str:
        pass

    def __str__(self) -> str:
        pass

    def _nb_units(self, domain: Optional[np.ndarray], samp_weight: np.ndarray) -> None:
        """Returns the number of units"""

        if domain is None:
            self.nb_units = len(samp_weight)
        elif domain is not None:
            keys, values = np.unique(domain, return_counts=True)
            self.nb_units = dict(zip(keys, values))

    @staticmethod
    def _deff_weight(samp_weight: np.ndarray) -> Number:
        """compute the design effect due to unequal weights -
        Page 71 of Valliant and Dever (2018)"""

        mean_w = np.mean(samp_weight)
        relvar_w = np.power(samp_weight - mean_w, 2) / mean_w**2

        return float(1 + np.mean(relvar_w))

    def calculate_deff_weight(
        self, samp_weight: Array, domain: Optional[np.ndarray] = None
    ) -> Union[DictStrNum, Number]:
        """Computes the design effect due to unequal weights.

        Args:
            samp_weight (Array):  array of the pre-adjustment sample weight. This vector
                should contains numeric values.
            domain (Optional[np.ndarray], optional): array indicating the normalization class
                for each sample unit. Defaults to None. Defaults to None.

        Returns:
            DictStrNum: dictionnary pairing the domains to the design effects due
                unequal weights.
        """

        samp_weight = formats.numpy_array(samp_weight)

        if domain is None:
            self.deff_weight = self._deff_weight(samp_weight)
            return self.deff_weight
        else:
            self.deff_weight = {}
            for d in np.unique(domain):
                self.deff_weight[d] = self._deff_weight(samp_weight[domain == d])
            return self.deff_weight

    @staticmethod
    def _norm_adjustment(
        samp_weight: np.ndarray,
        control: Number,
    ) -> tuple[np.ndarray, Number]:

        sum_weights = np.sum(samp_weight)
        adj_factor = float(control / sum_weights)

        return np.asarray(samp_weight * adj_factor), adj_factor

    @staticmethod
    def _response(resp_status: np.ndarray, resp_dict: Optional[dict[str, StringNumber]]) -> np.ndarray:

        resp_status = formats.numpy_array(resp_status)
        checks.assert_response_status(resp_status, resp_dict)

        if not np.isin(resp_status, ("in", "rr", "nr", "uk")).any() and resp_dict is not None:
            if "rr" not in resp_dict:
                raise ValueError("The response dictionary must contain the key 'rr'!")

            resp_code = np.repeat("  ", resp_status.size).astype(str)
            resp_code[resp_status == resp_dict["rr"]] = "rr"
            if "in" in resp_dict:
                resp_code[resp_status == resp_dict["in"]] = "in"
            if "nr" in resp_dict:
                resp_code[resp_status == resp_dict["nr"]] = "nr"
            if "uk" in resp_dict:
                resp_code[resp_status == resp_dict["uk"]] = "uk"
        else:
            resp_code = resp_status

        return resp_code

    @staticmethod
    def _adj_factor(
        samp_weight: np.ndarray, resp_code: np.ndarray, unknown_to_inelig: bool
    ) -> tuple[np.ndarray, Number]:

        rr_sample = resp_code == "rr"  # respondent
        in_sample = resp_code == "in"  # ineligible
        nr_sample = resp_code == "nr"  # nonrespondent
        uk_sample = resp_code == "uk"  # unknown

        # breakpoint()

        in_weights_sum = float(np.sum(samp_weight[in_sample]))
        rr_weights_sum = float(np.sum(samp_weight[rr_sample]))
        nr_weights_sum = float(np.sum(samp_weight[nr_sample]))
        uk_weights_sum = float(np.sum(samp_weight[uk_sample]))

        if unknown_to_inelig:
            adj_uk = (in_weights_sum + rr_weights_sum + nr_weights_sum + uk_weights_sum) / (
                in_weights_sum + rr_weights_sum + nr_weights_sum
            )
            adj_rr = (rr_weights_sum + nr_weights_sum) / rr_weights_sum
        else:
            adj_uk = 1
            adj_rr = (rr_weights_sum + nr_weights_sum + uk_weights_sum) / rr_weights_sum

        adj_factor = np.zeros(samp_weight.size)  # unknown and nonresponse will get 0 by default
        adj_factor[rr_sample] = adj_rr * adj_uk
        adj_factor[in_sample] = adj_uk

        return adj_factor, adj_rr

    def adjust(
        self,
        samp_weight: Array,
        adj_class: Array,
        resp_status: Array,
        resp_dict: Optional[Union[dict[str, StringNumber]]] = None,
        unknown_to_inelig: bool = True,
    ) -> np.ndarray:
        """adjusts sample weight to account for non-response.

        Args:
            samp_weight (np.ndarray): array of the pre-adjustment sample weight. This vector
                should contains numeric values.
            adj_class (np.ndarray): array indicating the adjustment class for each sample unit.
                The sample weight adjustments will be performed within the classes defined by this
                parameter.
            resp_status (np.ndarray): array indicating the eligibility and response status of the
                sample unit. Values of resp_status should inform on ineligible (in), respondent (rr), nonrespondent (nr), not known / unknown (uk). If the values of the parameter are not in ("in", "rr", "nr", "uk") then the resp_dict is required.
            resp_dict (Union[dict[str, StringNumber], None], optional): dictionnary providing the
                mapping between the values of resp_status and the ["in", "rr", "nr", "uk"].
                For example, if the response status are: 0 for ineligible, 1 for respondent,
                2 for nonrespondent, and 9 for unknown. Then the dictionary will be {"in": 0, "rr": 1, "nr": 2, "uk": 9}. If the response status variable has only values in ("in", "rr", "nr", "uk") then the dictionary is not needed. Optional parameter. Defaults to None.
            unknown_to_inelig (bool, optional): [description]. Defaults to True.

        Raises:
            AssertionError: raises an assertion error if adj_class is not a list, numpy array,
            or pandas dataframe/series.

        Returns:
            np.ndarray: array of the adjusted sample weights.
        """

        resp_code = self._response(formats.numpy_array(resp_status), resp_dict)
        samp_weight = formats.numpy_array(samp_weight)
        adjusted_weight = np.ones(samp_weight.size) * np.nan

        if adj_class is None:
            (
                adj_factor,
                self.adj_factor,
            ) = self._adj_factor(samp_weight, resp_code, unknown_to_inelig)
            adjusted_weight = adj_factor * samp_weight
        else:
            if isinstance(adj_class, list):
                adj_class = pd.DataFrame(np.column_stack(adj_class))
            elif isinstance(adj_class, np.ndarray):
                adj_class = pd.DataFrame(adj_class)
            elif not isinstance(adj_class, (pd.Series, pd.DataFrame, pl.Series, pl.DataFrame)):
                raise AssertionError(
                    "adj_class must be an numpy ndarray, a list of numpy ndarray or a pandas dataframe."
                )

            adj_array = formats.dataframe_to_array(adj_class)
            self.adj_factor = {}
            for c in np.unique(adj_array):
                samp_weight_c = samp_weight[adj_array == c]
                resp_code_c = resp_code[adj_array == c]
                adj_factor_c, self.adj_factor[c] = self._adj_factor(samp_weight_c, resp_code_c, unknown_to_inelig)
                adjusted_weight[adj_array == c] = adj_factor_c * samp_weight_c

        self.deff_weight = self.calculate_deff_weight(adjusted_weight)
        self.adj_method = "nonresponse"

        return np.asarray(adjusted_weight)

    @staticmethod
    def _core_matrix(
        samp_weight: np.ndarray,
        x: np.ndarray,
        x_weighted_total: np.ndarray,
        x_control: np.ndarray,
        scale: np.ndarray,
    ) -> np.ndarray:

        v_inv_d = np.diag(samp_weight / scale)
        core_matrix = np.dot(np.matmul(np.transpose(x), v_inv_d), x)
        if x.shape == (x.size,):
            core_factor = (x_control - x_weighted_total) / core_matrix
        else:
            core_factor = np.matmul(
                np.transpose(x_control - x_weighted_total),
                np.linalg.inv(core_matrix),
            )

        return np.asarray(core_factor)

    def normalize(
        self,
        samp_weight: Array,
        control: Optional[Union[DictStrNum, Number]] = None,
        domain: Optional[Array] = None,
    ) -> np.ndarray:
        """normalizes the sample weights to sum to a known constants or levels.

        Args:
            samp_weight (array) : array of the pre-adjustment sample weight. This vector should
                contains numeric values.
            control (int, float, dictionary) : a number or array of the level to calibrate the
                sum of the weights. Default is number of units by domain key or overall if domain
                is None. Defaults to None.
            domain (Optional[Array], optional) : array indicating the normalization class for each
                sample unit. Defaults to None.

        Returns:
            An arrays: the normalized sample weight.
        """

        samp_weight = formats.numpy_array(samp_weight)
        norm_weight = samp_weight.copy()

        if domain is not None:
            domain = formats.numpy_array(domain)
            keys = np.unique(
                pl.DataFrame(domain).filter(pl.col("column_0").is_not_null())["column_0"].to_numpy()
            )  # NoneType and Object types are problematic for np.unique()
            levels: np.ndarray = np.zeros(keys.size) * np.nan
            self.adj_factor = {}
            self.control = {}
            for k, key in enumerate(keys):
                weight_k = samp_weight[domain == key]
                if control is None:
                    levels[k] = np.sum(domain == key)
                elif control is not None and isinstance(control, dict):
                    levels[k] = control[key]
                elif isinstance(control, (float, int)):
                    levels[k] = control

                (
                    norm_weight[domain == key],
                    self.adj_factor[key],
                ) = self._norm_adjustment(weight_k, levels[k])
                self.control[key] = levels[k]
        else:
            if control is None:
                self.control = np.sum(samp_weight.size)
                norm_weight, self.adj_factor = self._norm_adjustment(samp_weight, self.control)
            elif isinstance(control, (int, float)):
                norm_weight, self.adj_factor = self._norm_adjustment(samp_weight, control)
                self.control = control

        self.adj_method = "normalization"

        return norm_weight

    def poststratify(
        self,
        samp_weight: Array,
        control: Optional[Union[DictStrNum, Number]] = None,
        factor: Optional[Union[DictStrNum, Number]] = None,
        domain: Optional[Array] = None,
    ) -> np.ndarray:
        """[summary]

        Args:
            samp_weight (Array): [description]
            control (Union[DictStrNum, Number, None], optional): a number or
                array of the level to calibrate the sum of the weights. Defaults to None.
            factor (Union[DictStrNum, Number, None], optional): adjustment factor.
                Defaults to None.
            domain (Optional[Array], optional): array indicating the normalization class for each
                sample unit. Defaults to None.

        Raises:
            AssertionError: raises an assertion error if both control and factor are not provided.
            ValueError: raises an error is control dictionary keys do not match domain's values.
            ValueError: raises an error is factor dictionary keys do not match domain's values.

        Returns:
            np.ndarray:  array of poststratified sample weights.
        """

        if control is None and factor is None:
            raise AssertionError("control or factor must be specified.")

        if isinstance(control, dict):
            # breakpoint()
            if (np.unique(domain) != np.unique(list(control.keys()))).any():
                raise ValueError("control dictionary keys do not match domain values.")

        if control is None and domain is not None:
            if isinstance(factor, dict) and (np.unique(domain) != np.unique(list(factor.keys()))).any():
                raise ValueError("factor dictionary keys do not match domain values.")

            sum_weight = float(np.sum(samp_weight))
            if isinstance(factor, dict):
                control = {}
                for d in np.unique(domain):
                    control[d] = sum_weight * factor[d]
            elif isinstance(factor, (int, float)):
                control = sum_weight * factor

        ps_weight = self.normalize(samp_weight, control, domain)
        self.adj_method = "poststratification"

        return ps_weight

    def _raked_wgt(
        self,
        samp_weight: np.ndarray,
        X: np.ndarray,
        control: Union[DictStrNum, None],
        domain: Optional[Array] = None,
    ) -> None:
        pass

    def rake(
        self,
        samp_weight: Array,
        margins: DictStrNum,
        control: Optional[DictStrNum] = None,
        factor: Optional[DictStrNum] = None,
        ll_bound: Optional[Union[DictStrNum, Number]] = None,
        up_bound: Optional[Union[DictStrNum, Number]] = None,
        tol: float = 1e-4,
        ctrl_tol: float = 1e-4,
        max_iter: int = 100,
        display_iter: bool = False,
    ) -> np.ndarray:

        samp_weight = formats.numpy_array(samp_weight)

        obs_tol = tol + 1
        iter = 0

        converged = False
        bounded = True

        while not (converged and bounded) and iter < max_iter:
            if display_iter:
                print(f"\nIteration {iter + 1}")

            if iter == 0:
                rk_wgt = samp_weight
                wgt_prev = samp_weight

            for margin in margins:
                domain = formats.numpy_array(margins[margin])
                if control is not None:
                    rk_wgt = self.poststratify(samp_weight=rk_wgt, control=control[margin], domain=domain)
                elif factor is not None:
                    rk_wgt = self.poststratify(samp_weight=rk_wgt, factor=factor[margin], domain=domain)
                else:
                    raise AssertionError("control or factor must be specified!")

            sum_wgt = {}
            sum_prev_wgt = {}
            for margin in margins:
                domain = formats.numpy_array(margins[margin])
                sum_wgt_domain = {}
                sum_prev_wgt_domain = {}
                for d in control[margin]:
                    sum_wgt_domain[d] = np.sum(rk_wgt[domain == d])
                    sum_prev_wgt_domain[d] = np.sum(wgt_prev[domain == d])
                sum_wgt[margin] = sum_wgt_domain
                sum_prev_wgt[margin] = sum_prev_wgt_domain

            # diff = {}
            max_diff = 0
            max_ctrl_diff = 0
            for margin in margins:
                if display_iter:
                    print(f"  Margin: {margin}")

                diff_margin = {}
                diff_ctrl_margin = {}
                for d in control[margin]:
                    diff_margin[d] = np.abs(sum_wgt[margin][d] - sum_prev_wgt[margin][d]) / sum_prev_wgt[margin][d]
                    diff_ctrl_margin[d] = np.abs(sum_wgt[margin][d] - control[margin][d]) / control[margin][d]
                    if display_iter:
                        print(f"    Difference against previous value for '{d}': {diff_margin[d]}")
                        print(f"    Difference against control value for '{d}': {diff_ctrl_margin[d]}")

                # diff[margin] = diff_margin
                max_diff = max(max_diff, max(diff_margin.values()))
                max_ctrl_diff = max(max_ctrl_diff, max(diff_ctrl_margin.values()))

            obs_tol = max_diff
            obs_ctrl_tol = max_ctrl_diff

            if obs_tol <= tol and obs_ctrl_tol <= ctrl_tol:
                converged = True

                if ll_bound is not None or up_bound is not None:
                    wgt_ratios = rk_wgt / samp_weight
                    min_ratio = np.min(wgt_ratios)
                    max_ratio = np.max(wgt_ratios)

                    if (
                        (
                            ll_bound is not None
                            and up_bound is not None
                            and (ll_bound <= min_ratio and max_ratio <= up_bound)
                        )
                        or (ll_bound is not None and ll_bound <= min_ratio)
                        or (up_bound is not None and up_bound >= max_ratio)
                    ):
                        bounded = True
                    else:
                        bounded = False
                else:
                    bounded = True

            wgt_prev = rk_wgt
            iter += 1

        self.adj_method = "raking"

        return rk_wgt

    @staticmethod
    def _calib_covariates(
        data: pd.DataFrame,
        x_cat: Optional[list[str]] = None,
        x_cont: Optional[list[str]] = None,
    ) -> tuple[np.ndarray, DictStrNum]:

        if not isinstance(data, pd.DataFrame) or data is None:
            raise ValueError("data must be a pandas dataframe.")

        if x_cat is None and x_cont is None:
            raise AssertionError("x_cat and/or x_cont must be specified.")
        else:
            if x_cat is not None:
                x_concat = formats.dataframe_to_array(data[x_cat])
                x_dummies = pd.get_dummies(x_concat)
                x_dict = formats.array_to_dict(x_concat)
                # x_dummies.insert(0, "intercept", 1)
            if x_cont is None and x_dummies is not None:
                x_array = x_dummies.astype("int")
            elif x_cont is not None and x_dict is not None:
                x_array = pd.concat([x_dummies, data[x_cont]], axis=1).astype("int")
                x_cont_dict: DictStrNum = {}
                nb_obs = data[x_cont].shape[0]
                for var in x_cont:
                    x_cont_dict[var] = nb_obs
                    x_dict.update(x_cont_dict)
            else:
                raise AssertionError

        return np.asarray(x_array.to_numpy()), x_dict

    def calib_covariates(
        self,
        data: pd.DataFrame,
        x_cat: Optional[list[str]] = None,
        x_cont: Optional[list[str]] = None,
        domain: Optional[list[str]] = None,
    ) -> tuple[np.ndarray, Union[DictStrNum, dict[StringNumber, DictStrNum]]]:
        """A utility function that creates an array of the calibration groups/domains and
        a dictionary pairing the domains with the control values.

        Args:
            data (pd.DataFrame): input pandas dataframe with the calibration's control data.
            x_cat (Optional[list[str]], optional): list of the names of the categorical control
                variables. Defaults to None.
            x_cont (Optional[list[str]], optional): list of the names of the continuous control
                variables. Defaults to None.
            domain (Optional[list[str]], optional): list of the names of the variables defining
                the normalization classes for each sample unit. Defaults to None.

        Raises:
            AssertionError: raises an assertion error if input data is not a pandas dataframe.

        Returns:
            tuple[np.ndarray, Union[DictStrNum, dict[StringNumber, DictStrNum]]]: a tuple of
            an array of the calibration domains and a dictionary pairing the domains with the
            control values.
        """

        if not isinstance(data, (pd.DataFrame, pd.Series)):
            raise AssertionError("data must be a pandas dataframe.")

        if isinstance(data[x_cat], pd.Series):
            nb_cols = (data[x_cat].drop_duplicates()).shape[0] + 1
        elif x_cont is None:
            nb_cols = (data[x_cat].drop_duplicates()).shape[0]
        else:
            nb_cols = (data[x_cat].drop_duplicates()).shape[0] + len(x_cont)

        x_dict: Union[DictStrNum, dict[StringNumber, DictStrNum]]
        if domain is None:
            x_array, x_dict = self._calib_covariates(data, x_cat, x_cont)
            for key in x_dict:
                x_dict[key] = np.nan
        else:
            x_dict2: dict[StringNumber, DictStrNum] = {}
            x_dict_d: DictStrNum
            x_array = np.zeros((data.shape[0], nb_cols))
            for d in np.unique(data[domain].values):
                (
                    x_array[data[domain] == d, :],
                    x_dict_d,
                ) = self._calib_covariates(data[data[domain] == d], x_cat, x_cont)
                for key in x_dict_d:
                    x_dict_d[key] = np.nan
                x_dict2[d] = x_dict_d
            x_dict = x_dict2

        if domain is None:
            return x_array, x_dict
        else:
            return x_array, x_dict

    def _calib_wgt(self, x: np.ndarray, core_factor: np.ndarray) -> np.ndarray:
        def _core_vector(x_i: np.ndarray, core_factor: np.ndarray) -> np.ndarray:
            return np.asarray(np.dot(core_factor, x_i))

        if x.shape == (x.size,):
            adj_factor = _core_vector(x, core_factor)
        else:
            adj_factor = np.apply_along_axis(_core_vector, axis=1, arr=x, core_factor=core_factor)

        return adj_factor

    def calibrate(
        self,
        samp_weight: Array,
        aux_vars: Array,
        control: Union[dict[StringNumber, Union[DictStrNum, Number]]],
        domain: Optional[Array] = None,
        scale: Union[Array, Number] = 1,
        bounded: bool = False,
        additive: bool = False,
    ) -> np.ndarray:
        """Calibrates the sample weights.

        Args:
            samp_weight (Array): array of sample weights.
            aux_vars (Array): array of auxiliary variables.
            control (Union[dict[StringNumber, Union[DictStrNum, Number]], None], optional):
                provides the controls by domain if applicable. Defaults to None.
            domain (Optional[Array], optional): Array indicating the normalization class for each
                sample unit. Defaults to None.
            scale (Union[Array, Number], optional): [description]. Defaults to 1.
            bounded (bool, optional): [description]. Defaults to False.
            additive (bool, optional): [description]. Defaults to False.

        Returns:
            np.ndarray: an array of the calibrated sample weights.
        """

        samp_weight = formats.numpy_array(samp_weight)
        aux_vars = formats.numpy_array(aux_vars)
        samp_size = samp_weight.size
        if domain is not None:
            domain = formats.numpy_array(domain)
        if isinstance(scale, (float, int)):
            scale = np.repeat(scale, samp_size)
        else:
            scale = formats.numpy_array(scale)
        if aux_vars.shape == (samp_size,):
            x_w = aux_vars * samp_weight
            one_dimension = True
        else:
            x_w = np.transpose(aux_vars) * samp_weight
            one_dimension = False

        if domain is None:
            if one_dimension:
                x_w_total = np.sum(x_w)
            else:
                x_w_total = np.sum(x_w, axis=1)
            core_factor = self._core_matrix(
                samp_weight=samp_weight,
                x=aux_vars,
                x_weighted_total=x_w_total,
                x_control=np.array(list(control.values())),
                scale=scale,
            )
            adj_factor = 1 + self._calib_wgt(aux_vars, core_factor) / scale
        else:
            domains = np.unique(domain)
            if additive:
                adj_factor = np.ones((samp_size, domains.size)) * np.nan
            else:
                adj_factor = np.ones(samp_size) * np.nan

            for k, d in enumerate(domains):
                if one_dimension:
                    x_w_total = np.sum(x_w)
                else:
                    x_w_total = np.sum(x_w, axis=1)

                x_d = aux_vars[domain == d]
                samp_weight_d = samp_weight[domain == d]
                if one_dimension:
                    x_w_total_d = np.sum(x_w[domain == d])
                else:
                    x_w_total_d = np.sum(np.transpose(x_w)[domain == d], axis=0)

                control_d = control.get(d)
                if isinstance(control_d, (int, float)):
                    control_d_values = [control_d]
                elif isinstance(control_d, dict):
                    control_d_values = list(control_d.values())
                else:
                    raise TypeError("Type of control not valid!")

                scale_d = scale[domain == d]
                if additive:
                    core_factor_d = self._core_matrix(
                        samp_weight=samp_weight,
                        x=aux_vars,
                        x_weighted_total=x_w_total_d,
                        x_control=np.array(control_d_values),
                        scale=scale,
                    )
                    adj_factor[:, k] = (domain == d) + self._calib_wgt(aux_vars, core_factor_d) / scale
                else:
                    core_factor_d = self._core_matrix(
                        samp_weight=samp_weight_d,
                        x=aux_vars[domain == d],
                        x_weighted_total=x_w_total_d,
                        x_control=np.array(control_d_values),
                        scale=scale_d,
                    )
                    adj_factor[domain == d] = 1 + self._calib_wgt(x_d, core_factor_d) / scale_d

        if additive:
            calib_weight = np.transpose(np.transpose(adj_factor) * samp_weight)
        else:
            calib_weight = samp_weight * adj_factor

        self.adj_method = "calibration"

        return calib_weight

    def trim(
        self,
    ) -> np.ndarray:

        pass



from __future__ import annotations

from typing import Optional, Union

from samplics.utils.types import Array, Number, Series, StringNumber


# import numpy as np
# import pandas as pd


class CalibrateEstimator:
    """Provides calibration based estimation"""

    def __init__(
        self, param: str, alpha: float = 0.05, rand_seed: Optional[int] = None
    ):
        pass

    def estimate(
        self,
        y: Array,
        calib_weight: Array,
        # calib_rep_weight: Optional[Array] = None,
        x: Optional[Array] = None,
        stratum: Optional[Series] = None,
        psu: Optional[Series] = None,
        ssu: Optional[Series] = None,
        domain: Optional[Series] = None,
        by: Optional[Series] = None,
        fpc: Union[dict[StringNumber, Number], Series, Number] = 1.0,
        deff: bool = False,
        coef_variation: bool = False,
        as_factor: bool = False,
        remove_nan: bool = False,
    ) -> None:
        pass
